functor Context (structure Absyn : ABSTRACT_SYNTAX) :>
    CONTEXT where Absyn = Absyn
=

struct

    structure Absyn = Absyn

    open Absyn
    open Types


    datatype typing =
	Variable of tp * support
      | Name of tp
	
    type ctx = typing Var.VarMap.map


    val empty = Var.VarMap.empty


    fun okTp (ctx, ltp) = 
	(case ltp of
	     (LUNIT _ | LNAT _ | LBOOL _ | LNATLIST _) => true
	   | LVAR (x, loc) => 
		 let val _ = Error.error ("unknown type: " ^ (Var.toString x), loc, NONE)
		 in false
		 end

	   | (LPROD (ltp1, ltp2, loc) |
	      LARROW (ltp1, ltp2, loc) | 
	      LNARROW (ltp1, ltp2, loc)) =>
		 okTp (ctx, ltp1) andalso okTp (ctx, ltp2)
	   | (LBOX(C, ltp1, loc) | LDIA(C, ltp1, loc)) => 
		 let fun okSupport (ctx, C) =
		     let fun unknown x = 
			 (case (Var.VarMap.find (ctx, x)) of
			      SOME (Name _) => false
			    | _ => let val info = "identifier "^(Var.toString x)^" is not a declared name"
				       val _ = Error.error ("ill-formed support", loc, 
							    SOME(info))
				   in true end)
		     in
			 not (List.exists unknown (Support.toList C))
		     end
		 in
		     okSupport (ctx, C) andalso okTp (ctx, ltp1) 
		 end)


    val find = Var.VarMap.find

    (***
     ** inserting functions do *not* check if 
     ** the inserted type is well-formed in the context 
     ***)

    fun insertName (ctx, v, A) = 
	Var.VarMap.insert (ctx, v, Name A)

    fun insertVar (ctx, v, A, C) =
	Var.VarMap.insert (ctx, v, Variable (A, C))


    fun shadowName (ctx, v, v') = 
	(case Var.VarMap.find (ctx, v) of
	     NONE => ctx
	   | SOME (Variable (tp, C)) => 
		 raise Error.Error "shadowing a variable instead of a name"
	   | SOME (Name A) => 
		     let val ctx' = #1 (Var.VarMap.remove (ctx, v))
			 val subst = Var.VarMap.insert (Var.VarMap.empty, v, Var v')
		         fun shadow (Variable (tp, C)) = Variable (substType (subst, tp),
								   substSupport (subst, C))
			   | shadow (Name tp) = Name (substType (subst, tp))
			 val ctx'' = Var.VarMap.map shadow ctx'
		     in
			 Var.VarMap.insert (ctx'', v', Name A)
		     end)
			 

end